from __future__ import annotations

from dataclasses import dataclass, field
from typing import Iterable, List, Sequence


@dataclass(frozen=True)
class TokenizationResult:
    token_ids: Sequence[int]
    token_strings: Sequence[str]
    total_tokens: int
    truncated: bool
    elapsed_ms: float

    def to_dict(self) -> dict[str, object]:
        return {
            "token_ids": list(self.token_ids),
            "token_strings": list(self.token_strings),
            "total_tokens": self.total_tokens,
            "truncated": self.truncated,
            "elapsed_ms": self.elapsed_ms,
        }


@dataclass(frozen=True)
class BenchmarkScenario:
    name: str
    input_source: str
    size_tokens: int
    expected_p50_ms: float
    notes: List[str] = field(default_factory=list)

    def describe(self) -> str:
        return (
            f"Scenario '{self.name}' from {self.input_source} expects "
            f"~{self.size_tokens} tokens @ p50 {self.expected_p50_ms}ms"
        )




from __future__ import annotations

import functools
import importlib
from concurrent.futures import ThreadPoolExecutor
from time import perf_counter
from typing import AbstractSet, Collection, Literal, Mapping, NoReturn, Optional, Sequence

from . import compat, errors, registry
from .configuration import TokenizerConfiguration, TokenizerRuntime
from .metrics import MetricsPayload, dispatch
from .results import TokenizationResult


_ALLOWED_SPECIAL_ALL = "all"
_DISALLOWED_SPECIAL_ALL = "all"


def _load_backend():
    return importlib.import_module("bpe_openai._bindings")


def _bpe_encode_bytes(mergeable_ranks: Mapping[bytes, int], data: bytes) -> list[int]:
    if not data:
        return []
    parts = [bytes([b]) for b in data]
    while True:
        min_rank = None
        min_idx = None
        for idx in range(len(parts) - 1):
            pair = parts[idx] + parts[idx + 1]
            rank = mergeable_ranks.get(pair)
            if rank is not None and (min_rank is None or rank < min_rank):
                min_rank = rank
                min_idx = idx
        if min_rank is None:
            break
        parts[min_idx : min_idx + 2] = [parts[min_idx] + parts[min_idx + 1]]
    return [mergeable_ranks[part] for part in parts]


class Encoding:
    """tiktoken-compatible encoding interface backed by the Rust tokenizer."""

    def __init__(
        self,
        *,
        name: str,
        model: str,
        runtime: TokenizerRuntime,
        backend,
        backend_version: str,
        pat_str: str,
        mergeable_ranks: dict[bytes, int],
        special_tokens: dict[str, int],
    ) -> None:
        self.name = name
        self._model = model
        self._runtime = runtime
        self._backend = backend
        self._backend_version = backend_version
        self._pat_str = pat_str
        # Copy to guard against accidental mutation of cached dictionaries.
        self._mergeable_ranks = dict(mergeable_ranks)
        self._special_tokens = dict(special_tokens)

        self._metrics_hook = None
        self._last_result: Optional[TokenizationResult] = None

        self._build_rank_tables()

    def __repr__(self) -> str:  # pragma: no cover - formatting helper
        return f"<Encoding {self.name!r}>"

    # ---------------------------------------------------------------------
    # Encoding helpers
    # ---------------------------------------------------------------------

    def encode_ordinary(self, text: str) -> list[int]:
        text = self._sanitize_text(text)
        tokens = self._encode_plain(text)
        self._check_chunk_limit(len(tokens))
        return tokens

    def encode(
        self,
        text: str,
        *,
        allowed_special: Literal["all"] | AbstractSet[str] = frozenset(),
        disallowed_special: Literal["all"] | Collection[str] = "all",
    ) -> list[int]:
        text = self._sanitize_text(text)
        allowed = self._normalize_allowed_special(allowed_special)
        disallowed = self._normalize_disallowed_special(allowed, disallowed_special)

        if disallowed:
            regex = _special_token_regex(disallowed)
            match = regex.search(text)
            if match:
                raise_disallowed_special_token(match.group())

        start = perf_counter()
        tokens = self._encode_with_special(text, allowed)
        elapsed_ms = (perf_counter() - start) * 1_000

        result = TokenizationResult(
            token_ids=tokens,
            token_strings=[],
            total_tokens=len(tokens),
            truncated=False,
            elapsed_ms=elapsed_ms,
        )
        self._last_result = result

        dispatch(
            self._metrics_hook,
            MetricsPayload(
                model=self._model,
                total_tokens=len(tokens),
                elapsed_ms=elapsed_ms,
                rust_backend_version=self._backend_version,
            ),
        )

        return tokens

    def encode_to_numpy(
        self,
        text: str,
        *,
        allowed_special: Literal["all"] | AbstractSet[str] = frozenset(),
        disallowed_special: Literal["all"] | Collection[str] = "all",
    ):
        tokens = self.encode(
            text,
            allowed_special=allowed_special,
            disallowed_special=disallowed_special,
        )
        import numpy as np  # Local import to avoid hard dependency unless needed.

        return np.asarray(tokens, dtype=np.uint32)

    def encode_ordinary_batch(
        self,
        text: Sequence[str],
        *,
        num_threads: int = 8,
    ) -> list[list[int]]:
        def worker(item: str) -> list[int]:
            tokens = self._encode_plain(self._sanitize_text(item))
            self._check_chunk_limit(len(tokens))
            return tokens

        with ThreadPoolExecutor(max_workers=num_threads) as pool:
            return list(pool.map(worker, text))

    def encode_batch(
        self,
        text: Sequence[str],
        *,
        num_threads: int = 8,
        allowed_special: Literal["all"] | AbstractSet[str] = frozenset(),
        disallowed_special: Literal["all"] | Collection[str] = "all",
    ) -> list[list[int]]:
        allowed = self._normalize_allowed_special(allowed_special)
        disallowed = self._normalize_disallowed_special(allowed, disallowed_special)

        def worker(item: str) -> list[int]:
            return self.encode(
                item,
                allowed_special=allowed,
                disallowed_special=disallowed,
            )

        with ThreadPoolExecutor(max_workers=num_threads) as pool:
            return list(pool.map(worker, text))

    def encode_with_unstable(
        self,
        text: str,
        *,
        allowed_special: Literal["all"] | AbstractSet[str] = frozenset(),
        disallowed_special: Literal["all"] | Collection[str] = "all",
    ) -> tuple[list[int], list[list[int]]]:
        raise NotImplementedError(
            "encode_with_unstable is not implemented in bpe-openai; tiktoken compatibility pending"
        )

    def encode_single_token(self, text_or_bytes: str | bytes) -> int:
        if isinstance(text_or_bytes, bytes):
            if text_or_bytes in self._mergeable_ranks:
                return self._mergeable_ranks[text_or_bytes]
            decoded = None
            try:
                decoded = text_or_bytes.decode("utf-8")
            except UnicodeDecodeError as exc:
                raise KeyError(f"{text_or_bytes!r} is not a valid token") from exc
            if decoded in self._special_tokens:
                return self._special_tokens[decoded]
            raise KeyError(f"{text_or_bytes!r} is not a valid token")

        text = self._sanitize_text(text_or_bytes)
        if text in self._special_tokens:
            return self._special_tokens[text]
        candidate = self._encode_plain(text)
        if len(candidate) != 1:
            raise KeyError(f"{text_or_bytes!r} does not correspond to a single token")
        return candidate[0]

    def register_special_tokens(self, mapping: Mapping[str, int]) -> None:
        raise NotImplementedError(
            "register_special_tokens is not supported; the underlying tokenizer has a fixed vocabulary"
        )

    # ---------------------------------------------------------------------
    # Decoding helpers
    # ---------------------------------------------------------------------

    def decode_bytes(self, tokens: Sequence[int]) -> bytes:
        chunks = []
        for token in tokens:
            chunk = self._token_to_bytes(token)
            chunks.append(chunk)
        return b"".join(chunks)

    def decode(self, tokens: Sequence[int], errors: str = "replace") -> str:
        return self.decode_bytes(tokens).decode("utf-8", errors=errors)

    def decode_single_token_bytes(self, token: int) -> bytes:
        return self._token_to_bytes(token)

    def decode_tokens_bytes(self, tokens: Sequence[int]) -> list[bytes]:
        return [self._token_to_bytes(token) for token in tokens]

    def decode_with_offsets(
        self,
        tokens: Sequence[int],
    ) -> tuple[str, list[int]]:
        token_bytes = [self._token_to_bytes(token) for token in tokens]

        offsets: list[int] = []
        text_len = 0
        for chunk in token_bytes:
            if chunk:
                offsets.append(max(0, text_len - (0x80 <= chunk[0] < 0xC0)))
                text_len += sum(1 for byte in chunk if not 0x80 <= byte < 0xC0)
            else:
                offsets.append(text_len)

        text = b"".join(token_bytes).decode("utf-8", errors="strict")
        return text, offsets

    def decode_batch(
        self,
        batch: Sequence[Sequence[int]],
        *,
        errors: str = "replace",
        num_threads: int = 8,
    ) -> list[str]:
        with ThreadPoolExecutor(max_workers=num_threads) as pool:
            return list(pool.map(lambda seq: self.decode(seq, errors=errors), batch))

    def decode_bytes_batch(
        self,
        batch: Sequence[Sequence[int]],
        *,
        num_threads: int = 8,
    ) -> list[bytes]:
        with ThreadPoolExecutor(max_workers=num_threads) as pool:
            return list(pool.map(self.decode_bytes, batch))

    # ---------------------------------------------------------------------
    # Misc helpers
    # ---------------------------------------------------------------------

    def token_byte_values(self) -> list[bytes]:
        return list(self._rank_to_bytes)

    @property
    def special_tokens_set(self) -> set[str]:
        return set(self._special_tokens)

    @property
    def n_vocab(self) -> int:
        return self.max_token_value + 1

    def is_special_token(self, token: int) -> bool:
        return token in self._special_token_values

    @property
    def last_result(self) -> Optional[TokenizationResult]:
        return self._last_result

    @property
    def eot_token(self) -> int:
        token = self._special_tokens.get("<|endoftext|>")
        if token is None:
            raise KeyError("<|endoftext|> token is not defined for this encoding")
        return token

    def set_metrics_hook(self, callback) -> None:
        self._metrics_hook = callback
        self._runtime.set_metrics_hook(callback)

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _normalize_allowed_special(
        self, allowed_special: Literal["all"] | AbstractSet[str]
    ) -> frozenset[str]:
        if allowed_special == _ALLOWED_SPECIAL_ALL:
            return frozenset(self.special_tokens_set)
        return frozenset(allowed_special or [])

    def _normalize_disallowed_special(
        self,
        allowed_special: frozenset[str],
        disallowed_special: Literal["all"] | Collection[str],
    ) -> frozenset[str]:
        if disallowed_special == _DISALLOWED_SPECIAL_ALL:
            return frozenset(self.special_tokens_set - allowed_special)
        return frozenset(disallowed_special or [])

    def _encode_with_special(self, text: str, allowed_special: frozenset[str]) -> list[int]:
        if not allowed_special:
            tokens = self._encode_plain(text)
            self._check_chunk_limit(len(tokens))
            return tokens

        regex = _special_token_regex(allowed_special)
        tokens: list[int] = []
        last_index = 0
        for match in regex.finditer(text):
            if match.start() > last_index:
                tokens.extend(self._encode_plain(text[last_index : match.start()]))
            tokens.append(self._special_tokens[match.group()])
            last_index = match.end()
        if last_index < len(text):
            tokens.extend(self._encode_plain(text[last_index:]))

        self._check_chunk_limit(len(tokens))
        return tokens

    def _encode_plain(self, text: str) -> list[int]:
        if not text:
            return []
        if len(text) >= 1_000_000:
            raise ValueError("Input too long to encode safely")
        try:
            tokens = self._backend.encode(text)
        except UnicodeEncodeError:
            sanitized = self._sanitize_text(text)
            tokens = self._backend.encode(sanitized)
        return list(tokens)

    def _check_chunk_limit(self, total_tokens: int) -> None:
        limit = self._runtime.chunk_limit
        if limit and total_tokens > limit:
            errors.raise_token_limit(total_tokens, limit)

    def _token_to_bytes(self, token: int) -> bytes:
        try:
            chunk = self._rank_to_bytes[token]
        except IndexError as exc:  # pragma: no cover - invalid token
            raise KeyError(f"Token id {token} is out of range for {self.name}") from exc
        if not chunk:
            raise KeyError(f"Token id {token} is not defined for {self.name}")
        return chunk

    def _build_rank_tables(self) -> None:
        max_mergeable = max(self._mergeable_ranks.values(), default=-1)
        max_special = max(self._special_tokens.values(), default=-1)
        self.max_token_value = max(max_mergeable, max_special)
        size = self.max_token_value + 1 if self.max_token_value >= 0 else 0
        self._rank_to_bytes: list[bytes] = [b""] * size
        for token_bytes, rank in self._mergeable_ranks.items():
            if rank >= len(self._rank_to_bytes):
                self._rank_to_bytes.extend([b""] * (rank + 1 - len(self._rank_to_bytes)))
            self._rank_to_bytes[rank] = token_bytes
        for token_str, value in self._special_tokens.items():
            if value >= len(self._rank_to_bytes):
                self._rank_to_bytes.extend([b""] * (value + 1 - len(self._rank_to_bytes)))
            self._rank_to_bytes[value] = token_str.encode("utf-8")
        self.max_token_value = len(self._rank_to_bytes) - 1 if self._rank_to_bytes else -1
        self._special_token_values = frozenset(self._special_tokens.values())

    def _encode_bytes(self, data: bytes) -> list[int]:
        return _bpe_encode_bytes(self._mergeable_ranks, data)

    def _sanitize_text(self, text: str) -> str:
        try:
            text.encode("utf-8")
            return text
        except UnicodeEncodeError:
            return text.encode("utf-16", "surrogatepass").decode("utf-16", "replace")


def build_encoding_from_model(model_name: str) -> Encoding:
    config = TokenizerConfiguration.for_model(model_name)
    runtime = TokenizerRuntime(config)

    encoding_name = config.encoding
    metadata = registry.ENCODING_CONSTRUCTORS[encoding_name]()
    bindings = _load_backend()
    backend = bindings.tokenizer_for_model(model_name)
    backend_version = getattr(bindings, "RUST_BACKEND_VERSION", "unknown")

    return Encoding(
        name=metadata["name"],
        model=model_name,
        runtime=runtime,
        backend=backend,
        backend_version=backend_version,
        pat_str=metadata["pat_str"],
        mergeable_ranks=metadata["mergeable_ranks"],
        special_tokens=metadata["special_tokens"],
    )


def build_encoding_from_name(encoding_name: str) -> Encoding:
    encoding_key = encoding_name.lower()
    if encoding_key not in registry.ENCODING_CONSTRUCTORS:
        raise errors.UnsupportedModelError(
            model_name=encoding_name,
            supported_models=list(registry.ENCODING_CONSTRUCTORS),
        )

    config = TokenizerConfiguration(
        model_name=encoding_key,
        encoding=encoding_key,
        special_tokens=registry.ENCODING_CONSTRUCTORS[encoding_key]()["special_tokens"],
    )
    config.validate()
    runtime = TokenizerRuntime(config)

    metadata = registry.ENCODING_CONSTRUCTORS[encoding_key]()
    bindings = _load_backend()
    backend = bindings.tokenizer_for_encoding(encoding_name)
    backend_version = getattr(bindings, "RUST_BACKEND_VERSION", "unknown")

    return Encoding(
        name=metadata["name"],
        model=encoding_name,
        runtime=runtime,
        backend=backend,
        backend_version=backend_version,
        pat_str=metadata["pat_str"],
        mergeable_ranks=metadata["mergeable_ranks"],
        special_tokens=metadata["special_tokens"],
    )


@functools.lru_cache(maxsize=128)
def _special_token_regex(tokens: AbstractSet[str]):
    if not tokens:
        raise ValueError("Cannot build regex for empty token set")
    try:
        import regex as re  # type: ignore
    except ImportError:  # pragma: no cover - fallback for environments without regex
        import re  # type: ignore
    pattern = "|".join(re.escape(token) for token in tokens)
    return re.compile(f"({pattern})")


def raise_disallowed_special_token(token: str) -> NoReturn:
    raise ValueError(
        f"Encountered text corresponding to disallowed special token {token!r}.\n"
        "If you want this text to be encoded as a special token, "
        f"pass it to `allowed_special`, e.g. `allowed_special={{{token!r}, ...}}`.\n"
        "If you want this text to be encoded as normal text, disable the check for this token "
        f"by passing `disallowed_special=(enc.special_tokens_set - {{{token!r}}})`.\n"
        "To disable this check for all special tokens, pass `disallowed_special=()`.\n"
    )



"""Tiktoken-compatible wrapper exposing the bpe-openai Rust tokenizer crate."""

from __future__ import annotations

from importlib import metadata
from pathlib import Path
from typing import Iterable

from . import compat
from .errors import (
    SpecialTokenCollisionError,
    TokenLimitError,
    TokenizerError,
    UnsupportedModelError,
)
from .tokenizer import Encoding, build_encoding_from_model, build_encoding_from_name

def _local_version() -> str:
    pyproject = Path(__file__).resolve().parents[1] / "pyproject.toml"
    try:
        content = pyproject.read_text(encoding="utf-8")
    except FileNotFoundError:  # pragma: no cover - defensive fallback
        return "0.0.0"

    for line in content.splitlines():
        stripped = line.strip()
        if stripped.startswith("version = "):
            _, value = stripped.split("=", 1)
            value = value.split("#", 1)[0].strip()
            return value.strip('"\'')
    return "0.0.0"


try:  # pragma: no cover - package metadata unavailable during local dev
    __version__ = metadata.version("bpe-openai")
except metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = _local_version()

__all__ = [
    "Encoding",
    "encoding_for_model",
    "encoding_name_for_model",
    "get_encoding",
    "list_supported_models",
    "list_encoding_names",
    "TokenizerError",
    "UnsupportedModelError",
    "SpecialTokenCollisionError",
    "TokenLimitError",
    "__version__",
]


def list_supported_models() -> list[str]:
    return compat.list_supported_models()


def encoding_for_model(model_name: str) -> Encoding:
    return build_encoding_from_model(model_name)


def get_encoding(encoding_name: str) -> Encoding:
    return build_encoding_from_name(encoding_name)


def list_encoding_names() -> list[str]:
    from . import registry

    return sorted(registry.ENCODING_CONSTRUCTORS)


def encoding_name_for_model(model_name: str) -> str:
    metadata = compat.get_metadata(model_name)
    return metadata.encoding



from __future__ import annotations

from dataclasses import dataclass
from importlib import import_module
from typing import Callable, Dict, Optional


@dataclass(frozen=True)
class MetricsPayload:
    model: str
    total_tokens: int
    elapsed_ms: float
    rust_backend_version: str = "unknown"

    def to_dict(self) -> Dict[str, object]:
        return {
            "model": self.model,
            "total_tokens": self.total_tokens,
            "elapsed_ms": self.elapsed_ms,
            "rust_backend_version": self.rust_backend_version,
        }


def dispatch(hook: Optional[Callable[[Dict[str, object]], None]], payload: MetricsPayload) -> None:
    if hook:
        data = payload.to_dict()
        if data["rust_backend_version"] == "unknown":
            try:
                bindings = import_module("bpe_openai._bindings")
                data["rust_backend_version"] = getattr(
                    bindings, "RUST_BACKEND_VERSION", "unknown"
                )
            except ModuleNotFoundError:  # pragma: no cover - best effort
                pass
        hook(data)



from __future__ import annotations

import base64
import gzip
from functools import lru_cache
from importlib import import_module
from pathlib import Path
from typing import Dict

try:  # Python >= 3.9
    from importlib import resources
except ImportError:  # pragma: no cover - fallback
    import importlib_resources as resources  # type: ignore


_DATA_PACKAGE = "bpe_openai.data"
_VENDOR_DATA_DIR = (
    Path(__file__).resolve().parents[2]
    / "vendor"
    / "rust-gems"
    / "crates"
    / "bpe-openai"
    / "data"
)


def _open_tokenizer_data(stem: str):
    resource_name = f"{stem}.tiktoken.gz"
    try:
        return resources.open_binary(_DATA_PACKAGE, resource_name)
    except (FileNotFoundError, ModuleNotFoundError, AttributeError):
        fallback = _VENDOR_DATA_DIR / resource_name
        if not fallback.is_file():  # pragma: no cover - defensive
            raise FileNotFoundError(f"Missing tokenizer data file: {fallback}")
        return open(fallback, "rb")


@lru_cache(maxsize=None)
def _load_mergeable_ranks(stem: str) -> Dict[bytes, int]:
    mergeable: Dict[bytes, int] = {}
    with _open_tokenizer_data(stem) as raw, gzip.open(raw, "rt", encoding="utf-8") as handle:
        for line in handle:
            line = line.strip()
            if not line:
                continue
            token_b64, rank_str = line.split(" ")
            mergeable[base64.b64decode(token_b64)] = int(rank_str)
    return mergeable


@lru_cache(maxsize=None)
def cl100k_base() -> dict:
    mergeable_ranks = _load_mergeable_ranks("cl100k_base")
    special_tokens = {
        "<|endoftext|>": 100_257,
        "<|fim_prefix|>": 100_258,
        "<|fim_middle|>": 100_259,
        "<|fim_suffix|>": 100_260,
        "<|endofprompt|>": 100_276,
    }
    pat_str = (
        r"""'(?i:[sdmt]|ll|ve|re)|[^\r\n\p{L}\p{N}]?+\p{L}++|\p{N}{1,3}+| ?[^\s\p{L}\p{N}]++[\r\n]*+|\s++$|\s*[\r\n]|\s+(?!\S)|\s"""
    )
    return {
        "name": "cl100k_base",
        "pat_str": pat_str,
        "mergeable_ranks": mergeable_ranks,
        "special_tokens": special_tokens,
    }


@lru_cache(maxsize=None)
def o200k_base() -> dict:
    mergeable_ranks = _load_mergeable_ranks("o200k_base")
    special_tokens = {
        "<|endoftext|>": 199_999,
        "<|endofprompt|>": 200_018,
    }
    pat_str = "|".join(
        [
            r"""[^\r\n\p{L}\p{N}]?[\p{Lu}\p{Lt}\p{Lm}\p{Lo}\p{M}]*[\p{Ll}\p{Lm}\p{Lo}\p{M}]+(?i:'s|'t|'re|'ve|'m|'ll|'d)?""",
            r"""[^\r\n\p{L}\p{N}]?[\p{Lu}\p{Lt}\p{Lm}\p{Lo}\p{M}]+[\p{Ll}\p{Lm}\p{Lo}\p{M}]*(?i:'s|'t|'re|'ve|'m|'ll|'d)?""",
            r"""\p{N}{1,3}""",
            r""" ?[^\s\p{L}\p{N}]+[\r\n/]*""",
            r"""\s*[\r\n]+""",
            r"""\s+(?!\S)""",
            r"""\s+""",
        ]
    )
    return {
        "name": "o200k_base",
        "pat_str": pat_str,
        "mergeable_ranks": mergeable_ranks,
        "special_tokens": special_tokens,
    }


@lru_cache(maxsize=None)
def voyage3_base() -> dict:
    mergeable_ranks = _load_mergeable_ranks("voyage3_base")
    special_tokens = {
        "<|endoftext|>": 160_255,
        "<|fim_prefix|>": 160_256,
        "<|fim_middle|>": 160_257,
        "<|fim_suffix|>": 160_258,
    }
    pat_str = (
        r"""'(?i:[sdmt]|ll|ve|re)|[^\r\n\p{L}\p{N}]?+\p{L}++|\p{N}| ?[^\s\p{L}\p{N}]++[\r\n]*+|\s++$|\s*[\r\n]|\s+(?!\S)|\s"""
    )
    return {
        "name": "voyage3_base",
        "pat_str": pat_str,
        "mergeable_ranks": mergeable_ranks,
        "special_tokens": special_tokens,
    }


ENCODING_CONSTRUCTORS = {
    "cl100k_base": cl100k_base,
    "o200k_base": o200k_base,
    "voyage3_base": voyage3_base,
}


@lru_cache(maxsize=None)
def _openai_public():
    try:
        return import_module("tiktoken_ext.openai_public")
    except ModuleNotFoundError as exc:  # pragma: no cover - depends on install
        raise RuntimeError(
            "tiktoken_ext is required to load legacy encodings (gpt2/r50k/p50k)."
        ) from exc


def _load_openai_public(name: str) -> dict:
    constructor = getattr(_openai_public(), name)
    return constructor()


@lru_cache(maxsize=None)
def gpt2() -> dict:
    return _load_openai_public("gpt2")


@lru_cache(maxsize=None)
def r50k_base() -> dict:
    return _load_openai_public("r50k_base")


@lru_cache(maxsize=None)
def p50k_base() -> dict:
    return _load_openai_public("p50k_base")


@lru_cache(maxsize=None)
def p50k_edit() -> dict:
    return _load_openai_public("p50k_edit")


ENCODING_CONSTRUCTORS.update(
    {
        "gpt2": gpt2,
        "r50k_base": r50k_base,
        "p50k_base": p50k_base,
        "p50k_edit": p50k_edit,
    }
)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Mapping, MutableMapping, Optional, Set

from . import compat, errors

DEFAULT_CHUNK_LIMIT = 200_000


SUPPORTED_MODELS: Mapping[str, str] = {
    name: metadata.encoding for name, metadata in compat.MODEL_REGISTRY.items()
}


@dataclass
class TokenizerConfiguration:
    """Configuration information required to build a TokenizerRuntime."""

    model_name: str
    encoding: str
    special_tokens: MutableMapping[str, int] = field(default_factory=dict)
    allowed_special: Set[str] = field(default_factory=set)
    disallowed_special: Set[str] = field(default_factory=set)
    chunk_limit: int = DEFAULT_CHUNK_LIMIT

    def validate(self) -> None:
        if self.chunk_limit <= 0:
            raise ValueError("chunk_limit must be positive")

        collisions = set(self.allowed_special) & set(self.disallowed_special)
        if collisions:
            raise ValueError(
                f"Special token cannot be both allowed and disallowed: {sorted(collisions)}"
            )

        duplicate_ids = self._find_duplicate_ids()
        if duplicate_ids:
            raise errors.SpecialTokenCollisionError(ids=sorted(duplicate_ids))

    def _find_duplicate_ids(self) -> Set[int]:
        seen: Set[int] = set()
        duplicates: Set[int] = set()
        for value in self.special_tokens.values():
            if value in seen:
                duplicates.add(value)
            else:
                seen.add(value)
        return duplicates

    @classmethod
    def for_model(cls, model_name: str) -> "TokenizerConfiguration":
        model = model_name.lower()
        if model not in SUPPORTED_MODELS:
            errors.raise_unsupported_model(model_name, SUPPORTED_MODELS.keys())

        metadata = compat.get_metadata(model)
        special_tokens = dict(metadata.special_tokens)
        config = cls(
            model_name=model,
            encoding=metadata.encoding,
            special_tokens=special_tokens,
            chunk_limit=metadata.chunk_limit,
        )
        config.validate()
        return config


class TokenizerRuntime:
    """Owns tokenizer state and exposes encode/decode operations."""

    def __init__(self, config: TokenizerConfiguration) -> None:
        self.config = config
        self._metrics_hook: Optional[Callable[[dict[str, object]], None]] = None

    def set_metrics_hook(self, callback: Optional[Callable[[dict[str, object]], None]]) -> None:
        self._metrics_hook = callback

    def emit_metrics(self, payload: dict[str, object]) -> None:
        if self._metrics_hook:
            self._metrics_hook(payload)

    @property
    def chunk_limit(self) -> int:
        return self.config.chunk_limit

    def register_special_tokens(self, mapping: Mapping[str, int]) -> None:
        raise NotImplementedError(
            "Custom special tokens are not supported; the underlying tokenizer is fixed."
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Mapping

from . import registry


@dataclass(frozen=True)
class ModelMetadata:
    encoding: str
    chunk_limit: int
    special_tokens: Mapping[str, int]


def _metadata(
    encoding: str,
    chunk_limit: int = 200_000,
    special_tokens: Mapping[str, int] | None = None,
) -> ModelMetadata:
    tokens = special_tokens
    if tokens is None:
        tokens = registry.ENCODING_CONSTRUCTORS[encoding]()["special_tokens"]
    return ModelMetadata(
        encoding=encoding,
        chunk_limit=chunk_limit,
        special_tokens=tokens,
    )


MODEL_REGISTRY: Dict[str, ModelMetadata] = {
    "cl100k_base": _metadata("cl100k_base"),
    "o200k_base": _metadata("o200k_base"),
    "voyage3_base": _metadata("voyage3_base"),
    "gpt-4o": _metadata("o200k_base"),
    "gpt-4o-mini": _metadata("o200k_base"),
    "gpt-4.1": _metadata("o200k_base"),
    "gpt-4.1-mini": _metadata("o200k_base"),
    "gpt-4o-128k": _metadata("o200k_base"),
    "gpt-4.1-128k": _metadata("o200k_base"),
    "voyage-3": _metadata("voyage3_base"),
    "gpt-3.5": _metadata("cl100k_base"),
    "gpt-3.5-turbo": _metadata("cl100k_base"),
    "gpt-3.5-turbo-0301": _metadata("cl100k_base"),
    "gpt-3.5-turbo-0401": _metadata("cl100k_base"),
    "gpt-3.5-turbo-0125": _metadata("cl100k_base"),
    "gpt-35-turbo": _metadata("cl100k_base"),
    "gpt-4": _metadata("cl100k_base"),
    "gpt-4-0314": _metadata("cl100k_base"),
    "gpt-4-0613": _metadata("cl100k_base"),
    "gpt-4-32k": _metadata("cl100k_base"),
    "gpt2": _metadata("gpt2", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "gpt-2": _metadata("gpt2", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "r50k_base": _metadata("r50k_base", chunk_limit=4_096, special_tokens={"<|endoftext|>": 50_256}),
    "p50k_base": _metadata("p50k_base", chunk_limit=4_096, special_tokens={"<|endoftext|>": 50_256}),
    "p50k_edit": _metadata(
        "p50k_edit",
        chunk_limit=4_096,
        special_tokens={
            "<|endoftext|>": 50_256,
            "<|fim_prefix|>": 50_281,
            "<|fim_middle|>": 50_282,
            "<|fim_suffix|>": 50_283,
        },
    ),
    "text-davinci-003": _metadata("p50k_base", chunk_limit=4_096, special_tokens={"<|endoftext|>": 50_256}),
    "text-davinci-002": _metadata("p50k_base", chunk_limit=4_096, special_tokens={"<|endoftext|>": 50_256}),
    "text-davinci-001": _metadata("r50k_base", chunk_limit=4_096, special_tokens={"<|endoftext|>": 50_256}),
    "text-curie-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-babbage-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-ada-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "davinci": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "curie": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "babbage": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "ada": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "code-davinci-002": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "code-davinci-001": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "code-cushman-002": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "code-cushman-001": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "davinci-codex": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "cushman-codex": _metadata("p50k_base", chunk_limit=8_000, special_tokens={"<|endoftext|>": 50_256}),
    "text-davinci-edit-001": _metadata(
        "p50k_edit",
        chunk_limit=8_000,
        special_tokens={
            "<|endoftext|>": 50_256,
            "<|fim_prefix|>": 50_281,
            "<|fim_middle|>": 50_282,
            "<|fim_suffix|>": 50_283,
        },
    ),
    "code-davinci-edit-001": _metadata(
        "p50k_edit",
        chunk_limit=8_000,
        special_tokens={
            "<|endoftext|>": 50_256,
            "<|fim_prefix|>": 50_281,
            "<|fim_middle|>": 50_282,
            "<|fim_suffix|>": 50_283,
        },
    ),
    "text-similarity-davinci-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-similarity-curie-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-similarity-babbage-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-similarity-ada-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-search-davinci-doc-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-search-curie-doc-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-search-babbage-doc-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "text-search-ada-doc-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "code-search-babbage-code-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
    "code-search-ada-code-001": _metadata("r50k_base", chunk_limit=2_048, special_tokens={"<|endoftext|>": 50_256}),
}


def list_supported_models() -> list[str]:
    return sorted(MODEL_REGISTRY)


def get_metadata(model_name: str) -> ModelMetadata:
    key = model_name.lower()
    try:
        return MODEL_REGISTRY[key]
    except KeyError as exc:  # pragma: no cover - handled by higher-level error
        raise KeyError(model_name) from exc


def get_default_special_tokens(encoding_name: str) -> Mapping[str, int]:
    metadata = MODEL_REGISTRY.get(encoding_name.lower())
    if metadata:
        return metadata.special_tokens
    # Fallback to primary model metadata if encoding maps elsewhere.
    for item in MODEL_REGISTRY.values():
        if item.encoding == encoding_name:
            return item.special_tokens
    return {}



from typing import List, Optional

class PyTokenizer:
    def encode(self, text: str) -> List[int]: ...
    def decode(self, tokens: List[int]) -> str: ...
    def count(self, text: str) -> int: ...
    def count_till_limit(self, text: str, token_limit: int) -> Optional[int]: ...
    def pretokenize(self, text: str) -> List[str]: ...

def tokenizer_for_model(model_name: str) -> PyTokenizer: ...
def tokenizer_for_encoding(encoding_name: str) -> PyTokenizer: ...
def supported_models() -> List[str]: ...
def supported_encodings() -> List[str]: ...

RUST_BACKEND_VERSION: str
PYTHON_API_VERSION: str



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Sequence


class TokenizerError(RuntimeError):
    """Base class for wrapper-specific errors."""


@dataclass
class UnsupportedModelError(TokenizerError):
    model_name: str
    supported_models: Sequence[str]

    def __post_init__(self) -> None:
        supported = ", ".join(sorted(self.supported_models))
        super().__init__(
            f"Model '{self.model_name}' is not supported. Supported models: {supported}"
        )


@dataclass
class SpecialTokenCollisionError(TokenizerError):
    tokens: Sequence[str] | None = None
    ids: Sequence[int] | None = None

    def __post_init__(self) -> None:
        parts: list[str] = []
        if self.tokens:
            parts.append(f"tokens {sorted(self.tokens)}")
        if self.ids:
            parts.append(f"ids {sorted(self.ids)}")
        detail = " and ".join(parts) if parts else "unknown collision"
        super().__init__(f"Special token collision detected for {detail}")


@dataclass
class TokenLimitError(ValueError, TokenizerError):
    token_count: int
    chunk_limit: int

    def __post_init__(self) -> None:
        super().__init__(
            f"Input produces {self.token_count} tokens which exceeds chunk limit {self.chunk_limit}"
        )


def raise_unsupported_model(model_name: str, supported: Iterable[str]) -> None:
    raise UnsupportedModelError(model_name=model_name, supported_models=tuple(sorted(supported)))


def raise_token_limit(token_count: int, chunk_limit: int) -> None:
    raise TokenLimitError(token_count=token_count, chunk_limit=chunk_limit)
